
import java.awt.*;

/**
 *
 * This class draws the Huffman coding (sub)tree(s). To simplify the recursive 
 * drawing algorithm, it uses a grid layout for the trees. Each node is pointed 
 * on a grid point, and occupies a grid column of its own. The variables ldist
 * and rdist in the class Huffman.TreeNode give the neccessary information, 
 * namely, at which distance the right and the left child of a node have to 
 * drawn - and that is done recursively. See the methods paint and 
 * recChildrenDraw. Each leave and node is labeled by the methods labelLeave 
 * and labelNode, respectively. The occurence of a node or a leave is written 
 * in the center of the filled circle representing it in the tree. The character
 * of a leave is written below it as 'a', for example. The '' are used so that
 * a space can be recognized as ' '.
 * <p>
 * Further, a HuffComponent is to be inserted into a ScrollPane object. This is 
 * because the tree quickly becomes to big to be legibly drawn by itself. 
 * <p>
 * (c) R.Reichert & C.Oehler 1998, minor revision C. Sigg 2004 *
 */
public class HuffComponent extends Component {

	private class GridStructure {
		// 
		// A very rudimentary grid structure. Filled circles can be drawn on
		// grid points,
		// and they can be labeled in their center or below them. Further, the
		// circles
		// can be connected. The methods use the Graphics objects passed
		// as parameters. They do NOT change the color! The caller of these
		// methods has
		// to make sure the colors make sense (ie not writing in the center of a
		// filled
		// circle with that circle's color).
		// Note that on the left and on the top of the grid, half a column and
		// half a
		// row, respectively, are left empty. Looks nicer.
		//
		public int radius; // the radius of the nodes

		public int width; // the width of the grid fields

		public int height; // the height of the grid fields

		public GridStructure(int r, int w, int h) {
			radius = r;
			width = w;
			height = h;
		} /* GridStructure */

		public void circleAt(Graphics g, int gx, int gy) {
			g.fillOval(width / 2 + gx * width - radius / 2, height / 2 + gy
					* height - radius / 2, radius, radius);
		} /* circleAt */

		public void connectCircles(Graphics g, int gx1, int gy1, int gx2,
				int gy2) {
			g.drawLine(width / 2 + gx1 * width, height / 2 + gy1 * height,
					width / 2 + gx2 * width, height / 2 + gy2 * height);
		} /* connectCircles */

		public void drawStringCenter(Graphics g, String s, int gx, int gy) {
			int strHeight = g.getFontMetrics().getHeight();
			int strWidth = g.getFontMetrics().stringWidth(s);
			g.drawString(s, width / 2 + gx * width - strWidth / 2, height / 2
					+ gy * height + strHeight / 2);
		} /* drawStringCenter */

		public void drawStringBelow(Graphics g, String s, int gx, int gy) {
			int strHeight = g.getFontMetrics().getHeight();
			int strWidth = g.getFontMetrics().stringWidth(s);
			g.drawString(s, width / 2 + gx * width - strWidth / 2, height / 2
					+ gy * height + radius);
		} /* drawStringBelow */

	} /* GridStructure */

	private Huffman huff = null; // connection to a Huffman object

	private GridStructure grid; // the grid

	private Graphics g; // for convenience - less parameter passing

	private int width = 0; // the total width of this component - tree(s)
						   // dependent

	private int height = 0; // the total height of this component - dito

	private Color labelColor; // color of the leaves' labels

	private Color circleColor; // color of the filled circles

	public HuffComponent(Color labelColor, Color circleColor) {
		this.labelColor = labelColor;
		this.circleColor = circleColor;
		grid = new GridStructure(30 /* radius */, 18 /* width */, 35 /* height */);
		// hard-coded - could be made to depend on the trees...
	} /* HuffComponent */

	public static String formatFloat(float x) {
		if (x < 1)
			if (String.valueOf(x).length() > 3)
				return String.valueOf(x).substring(1, 4);
			else
				// a number like 0.1, 0.2 ...
				return String.valueOf(x).substring(1, 3);
		else
			return String.valueOf(x).substring(0, 1);
	} /* formatFloat */

	private void labelLeave(TreeNode leave, int curX, int curY) {
		String label = new String("'" + leave.theChar + "'");
		g.setColor(labelColor);
		grid.drawStringBelow(g, label, curX, curY);
		labelNode(leave, curX, curY);
	} /* labelLeave */

	private void labelNode(TreeNode node, int curX, int curY) {
		String label = formatFloat(node.prob);
		g.setColor(getBackground());
		grid.drawStringCenter(g, label, curX, curY);
		g.setColor(circleColor);
	} /* labelNode */

	private void recChildrenDraw(TreeNode parent, int curX, int curY) {
		// because we wish to connect a parent node to its children,
		// we proceed as follows: as a precondition, the parent's circle
		// is already drawn. So we draw the childrens' circles, connect
		// them to the parent circle, and then label the parent node.
		if (parent.left != null) {
			// 'parent' is a node: recursively draw left and right subtrees.
			// And since parent is a node, call labelNode
			int oldX = curX, oldY = curY;

			curX = oldX - parent.left.rdist;
			curY = oldY + 1;
			grid.circleAt(g, curX, curY);
			grid.connectCircles(g, oldX, oldY, curX, curY);
			recChildrenDraw(parent.left, curX, curY);

			curX = oldX + parent.right.ldist;
			curY = oldY + 1;
			grid.circleAt(g, curX, curY);
			grid.connectCircles(g, oldX, oldY, curX, curY);
			recChildrenDraw(parent.right, curX, curY);

			labelNode(parent, oldX, oldY);
		} else
			// 'parent' is a leave, its circle is already drawn: just label it
			labelLeave(parent, curX, curY);
	} /* recChildrenDraw */

	public void paint(Graphics g) {
		// if connected to some Huffman object, proceed through its list of
		// subtree roots. Draw the circle for each root, then call
		// recChildrenDraw
		// to complete the subtree
		if (huff != null) {
			this.g = g;
			g.setColor(circleColor);

			SortedList list = huff.getList();
			TreeNode node = (TreeNode) list.getFirst();
			int curX = 0, curY = 0, lastRdist = 0;

			while (node != null) {
				curX = curX + lastRdist + node.ldist;

				grid.circleAt(g, curX, curY);
				recChildrenDraw(node, curX, curY);

				lastRdist = node.rdist;
				node = (TreeNode) list.getNext();
			}
		} else
			super.paint(g);
	} /* paint */

	private int getTotalWidth(Huffman huff) {
		// compute the total grid width of the connected Huffman object.
		// This total grid width is NOT just the number of nodes in the
		// Huffman object (even though each node occupies one grid column):
		// we leave an empty column between the subtrees - looks nicer.
		// Used in update
		TreeNode current = (TreeNode) huff.getList().getFirst();
		int width = 0;

		while (current != null) {
			width += ((TreeNode) current).ldist + ((TreeNode) current).rdist;
			current = (TreeNode) huff.getList().getNext();
		}

		return width;
	} /* getTotalWidth */

	private int getTotalHeight(Huffman huff) {
		// compute the total grid height of the connected Huffman object.
		// This is just the maximal height of the subtrees. Used in update
		TreeNode current = (TreeNode) huff.getList().getFirst();
		int height = 0;

		while (current != null) {
			if (((TreeNode) current).height > height)
				height = ((TreeNode) current).height;
			current = (TreeNode) huff.getList().getNext();
		}

		return height;
	} /* getTotalHeight */

	public void update() {
		// compute the size of the component and set it.
		// Call the scroll pane's owner to get the scroll pane re-layoutet -
		// it may be neccessary to add or remove scrollbars. This is only done
		// if the scroll pane's owner lays out itself
		if (huff != null) {
			width = (getTotalWidth(huff) + 1) * grid.width;
			// +1 because of the grid is inset by half a grid.width
			height = (getTotalHeight(huff) + 2) * grid.height;
			// +2 because of the half grid.height inset plus label
			setSize(width, height);
		} else
			setSize(width = 0, height = 0);

		getParent().getParent().doLayout();
		repaint(); // in case the layout does not change, call repaint manually
	} /* update */

	public void setHuffman(Huffman huff) {
		this.huff = huff; // connect
	} /* setHuffman */

	public Dimension getPreferredSize() {
		return new Dimension(width, height); // this size would fit us
	} /* getPreferredSize */

} /* HuffComponent */

