
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * Huffman takes a string and runs the Huffman encoding algorithm on it. The
 * main goal is to construct the Huffman encoding tree for the input. Since this
 * is for tutorial purposes, efficiency is not our greatest concern. The class
 * offers the possibility of either constructing the tree in one rush, or in
 * individual steps, which allows the user to observe how the tree is built from
 * the bottom up.
 * <p>
 * Remember that the Huffman algorithm works as follows:
 * <ol>
 * <li>count the occurences of each character in the input</li>
 * <li>build a list where the items represent the characters, ordering the
 * items by occurences of the characters, ascending</li>
 * <li>while there is more than one item in the list: remove the first two
 * items from the list; build a new item: a node of the Huffman tree. Its
 * children are the items just removed from the list. Its "occurence" is the sum
 * of the occurences of its children; insert this new item in the sorted list
 * </li>
 * </ol>
 * At the end, there is only one item left in the list: the root of the Huffman
 * tree. This is exactly the algorithm implemented by the class Huffman. The
 * class TreeNode is used to store the nodes and leaves of the Huffman tree. It
 * takes care of building new nodes from two old ones.
 * </p>
 * <p>
 * (c) R.Reichert & C.Oehler 1998, minor revision by C. Sigg, 2004
 * </p>
 */
public class Huffman {

	public static final int HASHTABLE_SIZE = 113;

	private float entropy = 0; // the entropy of the input

	private float avgLength = 0; // the average length of the code

	private SortedList list = new SortedList(); // the list with the subtree

	// roots

	private String input; // user's input

	private int N = 0; // the number of diff. chars in input

	/**
	 * Build the list with the leaves of the Huffman tree, calculate the entropy
	 * while doing so. Remember that the entropy is calculated as: <br>
	 * entropy (input) = sum_{char c in input} prob(c) * log_2 (prob (c)), <br>
	 * where "_" denotes subscript and where prob denotes the "probability" of
	 * c, ie. the number of its occurences divided by the length of the input.
	 * 
	 * @param theInput
	 */
	public Huffman(String theInput) {
		input = new String(theInput);
		int length = input.length();
		Hashtable counts = new Hashtable(HASHTABLE_SIZE);
		float log2 = (float) Math.log(2);

		for (int i = 0; i < length; i++) {
			char c = input.charAt(i);
			TreeNode count = (TreeNode) counts.get(new Character(c));
			if (count != null)
				count.count++;
			else {
				counts.put(new Character(c),
						new TreeNode(c, 1, 0.0f /* dummy */));
				N++;
			}
		}

		Enumeration e = counts.elements();
		while (e.hasMoreElements()) {
			TreeNode count = (TreeNode) e.nextElement();
			float prob = (float) count.count / length;
			if (prob == 1)
				avgLength = 1;
			list.insert(new TreeNode(count.theChar, count.count, prob), false);
			entropy -= prob * Math.log(prob) / log2;
		}
	} /* Huffman */

	/**
	 * Implements Huffmans algorithm, removing two nodes, constructing a new
	 * node, and inserting this new node. The only "trick" here is the
	 * calculation of the average code length, which normally is computed as:
	 * <br>
	 * avg code length = sum_{ codeword c_i } prob (c_i) * length (c_i), ie, as
	 * the weighted sum of the code word lengths. However, it can also be
	 * computed as: <br>
	 * avg code length = sum_{ n inner node of tree } prob(n); ie as the sum of
	 * the probabilities of the inner nodes of the tree. This formula can be
	 * easily be computed while constructing the tree: each time a new node (ie
	 * an inner node) is constructed, add its probability to avgLength.
	 */
	public void compress() {
		if (list.count() > 0) {
			TreeNode first = (TreeNode) list.removeFirst();
			while (!list.empty()) {
				TreeNode second = (TreeNode) list.removeFirst();
				list.insert(new TreeNode(first, second), false);
				avgLength += first.prob + second.prob;
				first = (TreeNode) list.removeFirst();
			}
			list.insert(first, false);
		}
	} /* createTree */

	/**
	 * Does only one step of the Huffman algorithm. Same logic applies as in
	 * createTree.
	 * 
	 * @return
	 */
	public boolean oneStep() {
		if (list.count() > 1) {
			TreeNode first = (TreeNode) list.removeFirst();
			TreeNode second = (TreeNode) list.removeFirst();
			list.insert(new TreeNode(first, second), false);
			avgLength += first.prob + second.prob;
		}

		return list.count() == 1;
	} /* oneStep */

	/**
	 * For each leaf in the Huffman tree, store a pointer to it in the
	 * hashtable. Used only by output.
	 * 
	 * @param node
	 * @param codes
	 */
	private void makeCodesTable(TreeNode node, Hashtable codes) {
		if (node != null)
			if (node.left == null && node.right == null)
				codes.put(new Character(node.theChar), node);
			else {
				makeCodesTable(node.left, codes);
				makeCodesTable(node.right, codes);
			}
	} /* makeCodesTable */

	/**
	 * Construct the binary output. Call makeCodesTable to get a pointer to the
	 * leaves of the Huffman tree for each character in the input, then traverse
	 * the input char wise, appending the output for each char.
	 * 
	 * @return
	 */
	public String output() {
		if (list.count() == 1) {
			Hashtable codes = new Hashtable(HASHTABLE_SIZE);
			StringBuffer theOutput = new StringBuffer();

			makeCodesTable((TreeNode) list.getFirst(), codes);
			for (int i = 0; i < input.length(); i++) {
				char c = input.charAt(i);
				TreeNode charNode = (TreeNode) codes.get(new Character(c));
				theOutput.append(charNode.code + " ");
			}
			return new String(theOutput);
		} else
			return null;
	} /* output */

	/**
	 * Put the leaves of the tree 'node' in the array 'leaves', where the
	 * current node is to be put at the index 'index'. Used only in getNodesArray.
	 * 
	 * @param node
	 * @param leaves
	 * @param index
	 * @return
	 */
	private int makeLeavesArray(TreeNode node, TreeNode[] leaves, int index) {
		if (node != null)
			if (node.left == null && node.right == null) {
				leaves[index++] = node;
			} else {
				index = makeLeavesArray(node.left, leaves, index);
				index = makeLeavesArray(node.right, leaves, index);
			}
		return index;
	} /* makeLeavesArray */

	/**
	 * 	Put the leaves of all subtrees in the current list in an array.
	 *	This is done by calling makeLeavesArray for each subtree root in the
	 *	list.
	 *	The leave list can then be used to output the codes for each char,
	 *	for example.
	 * 
	 * @return
	 */
	public TreeNode[] getNodesArray() {
		if (list.count() > 0) {
			TreeNode[] nodes = new TreeNode[N];
			TreeNode current = (TreeNode) list.getFirst();
			int index = 0;

			while (current != null) {
				index = makeLeavesArray(current, nodes, index);
				current = (TreeNode) list.getNext();
			}
			return nodes;
		} else
			return null;
	} /* getNodesArray */

	public SortedList getList() {
		return list;
	} /* getList */

	public float getEntropy() {
		return entropy;
	} /* getEntropy */

	public float getAvgLength() {
		return avgLength;
	} /* getAvgLength */

} /* Huffman */

