/** 
 * TreeNode represents either a leave or a node in the Huffman tree.
 * It stores all the relevant information, and takes care of
 * maintaining the information when two leaves or nodes are being
 * "merged"
 * into a new node.
 *
 */
public class TreeNode implements Sortable {

	public char theChar; // the character represented by this node

	public int count; // the number of occurences of this char

	public float prob; // the probabilty of this char

	public int height; // the height of this subtree

	public TreeNode left; // the left child of this node - if any

	public TreeNode right; // the right child of this node - if any

	public StringBuffer code; // the code computed for this char

	public int ldist, rdist; // used by HuffComponent for the graphical

	// layout:

	// ldist indicates the size of the left subtree (including the root),
	// rdist indicates the size of the right subtree (including the root).
	// In the "grid layout" used by HuffComponent, each tree node occupies
	// one
	// "grid column".
	// Therefore, ldist and rdist also indicate at which "grid distance"
	// this subtree has to be drawn on the RIGHT or LEFT hand side of
	// its parent node, respectively.

	public TreeNode(char aChar, int aCount, float aProb) {
		// Initializes a leave of the Huffman tree - straightforward.
		count = aCount;
		prob = aProb;
		theChar = aChar;
		height = 0;
		left = null;
		right = null;
		ldist = 1;
		rdist = 1;
		if (prob == 1)
			code = new StringBuffer("1");
		else
			code = new StringBuffer("");
		// catch the very special case where there is only
		// a single char to be encoded
	} /* TreeNode */

	public TreeNode(TreeNode aLeft, TreeNode aRight) {
		// Merge two Huffman trees and maintain the information about the
		// new node.
		// Precondition: aLeft != null && aRight != null. This is
		// guaranteed by Huffman algorithm.
		left = aLeft;
		left.addChar('0'); // add prefix to the left subtree
		right = aRight;
		right.addChar('1'); // and to the right subtree

		theChar = 0; // nodes don't represent any character

		count = left.count + right.count;
		prob = left.prob + right.prob;
		height = (left.height > right.height) ? left.height + 1
				: right.height + 1;

		ldist = left.ldist + left.rdist;
		rdist = right.ldist + right.rdist;
	} /* TreeNode */

	protected void addChar(char c) {
		// Add c as a prefix to the code of the leaves.
		if (left == null && right == null)
			code.insert(0, c);
		else {
			left.addChar(c);
			right.addChar(c);
		}
	} /* addChar */

	public int compareTo(Sortable s) {
		// Compare the occurences of two nodes - ascending!
		// It was asked of us that leaves of the same
		// probability be oirdered alphabetically... here you go...
		TreeNode other = (TreeNode) s;
		if (left == null && other.left == null && // both are leaves
				count == other.count) // both have same prob
			return theChar - other.theChar;
		else
			return (count - other.count);
	} /* compareTo */

} /* TreeNode */

